function [hme] = hmeFit(hme,x,y,x2,y2,iter,tol,split,vis,prefix)
% [hme] = hmeFit(hme,x,y,iter,tol,split,vis)
%
% Fit a k-class HME model.
%
% INPUTS
%	hme	HME model with params initialized
%	x	dxn matrix of samples
%	y	kxn matrix of class assignments
%   x2  dxn2
%   y2  kxn2
%	[iter]	Number of EM iterations.  If a vector, then perform
%		at least min(iter) and at most max(iter) iterations.
%		Default value is [20 50].  A single number is valid.
%	[tol]	Stop EM when log likelihood increases by a fraction
%		less than tol, i.e. (old-new)/old < tol.  
%		Used when a range of iterations is given.  
%		Default value is 1e-4. 
%	[split] Value in (0,1) giving fraction of data to use as
%		the training set.  The remaining data is used as
%		the test set.  Default value is 0.5.
%	[vis]	Visualization level:
%		  0 = none [default]
%		  1 = plot log likelihood vs. iteration
%		  2 = image model details
%	[prefix] Prefix for visualization output files.
%
% OUTPUTS
%	hme	HME model with params fitted
%
% Calls logistK, logistK_eval.
%
% David Martin <dmartin@eecs.berkeley.edu> 
% Charless Fowlkes <fowlkes@eecs.berkeley.edu>
% May 7, 2002

error(nargchk(3,10,nargin));

if nargin < 6, iter = [10 50]; end
if nargin < 7, tol = 1e-4; end
if nargin < 8, split = 0.5; end
if nargin < 9, vis = 0; end
if nargin < 10, prefix = ''; end

[d,n] = size(x); 
[k,n] = size(y);

if isempty(x2) % Nao foi fornecido conjunto de validacao
    % split the samples into training and test sets
    n1 = round(n*split); n2 = n - n1;
    if n1 <= 0, error('there is no training data!'); end
    perm = randperm(n);
    ind1 = perm(1:n1); ind2 = perm(n1+1:n);
    x1 = x(:,ind1); x2 = x(:,ind2);
    y1 = y(:,ind1); y2 = y(:,ind2);
else
    x1=x;
    y1=y;
    n1=size(x1,2);
    n2=size(x2,2);
end

L1 = zeros(1,max(iter));
L2 = zeros(1,max(iter));

if vis > 0,
  h = figure;
  set(h,'DoubleBuffer','on');
end

lli = -inf;
for i = 1:max(iter),
  % EM using the training set
  hme = Estep(hme,x1,y1);
  hme = Mstep(hme,x1,y1,ones(1,n1));
  
  if vis > 1,
    % visualize the current state of the model
    hmeVis(hme,k,[0 1],[0 1],x1,prefix);
  end

  if n2 > 0 & (i > min(iter) | vis > 0),
    % evaluate the model using the test set
    [post2,lik2,lli2] = hmeEval(hme,k,x2,y2);
    L2(i) = lli2/n2;
    lli_prev = lli; lli = lli2/n2;
    disp(sprintf('hme iter=%d lli=%g',i,lli));
  else
    disp(sprintf('hme iter=%d',i));
  end

  if vis > 0,
    % evaluate the model using the training set
    [post1,lik1,lli1] = hmeEval(hme,k,x1,y1);
    L1(i) = lli1/n1;
    % plot log likelihood for training,test sets over time
    figure(h); hold off; plot(L1(1:i),'b-o');
    if n2 > 0, hold on; plot(L2(1:i),'r-o'); end
    if n2 > 0,
%      axis([1 max(iter) 1.01*min([L1(1:i) L2(1:i)]) 0.99*max([L1(1:i) L2(1:i)])]);
%      axis([1 max(iter) 1.01*min([L1(1:i) L2(1:i)]) 0.99*max([L1(1:i) L2(1:i)])]);
    else
 %     axis([1 max(iter) 1.01*min(L1(1:i)) 0.99*max(L1(1:i))]);
    end
    xlabel('iteration'); ylabel('mean log likelihood'); 
    if n2 > 0, legend('training set','test set',4); end
    print(h,'-depsc',[prefix 'lli']);
  end

  if n2 > 0 & i > min(iter),
    % stop iterating if the log-likelihood on the test set
    % decreases by a fraction less than tol
    if (lli_prev-lli)/lli < tol, break, end
  end
  
  figure(4)
  clf
  plot(x1,post1,'*')
  hold on
  plot(x1,y1,'r*')
  
  pause(2)

end

if vis == 1,
  % visualize the current state of the model
  hmeVis(hme,k,[0 1],[0 1],x1,prefix);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% requires hme.param in all nodes
% creates hme.clik in internal nodes
% creates hme.lik in all nodes
function [hme] = Estep(hme,x,y)

  if hme.leaf,
   
    % compute likelihood of expert
    switch lower (hme.type)
        case 'linear'
            [post,lik] = linearK_eval(hme.param,hme.sig,x,y);
        case 'mlp'
            [post,lik] = mlpk_eval(hme.param,hme.sig,x,y);
        case 'svm'
            [post,lik] = svmK_eval(hme.param,hme.sig,x,y);
        case 'lssvm'
            [post,lik] = lssvmK_eval(hme.param,hme.sig,x,y);            
    end
    hme.lik = lik;
    
  else
    
    % compute likelihood of children
    for i = 1:length(hme.children),
      hme.children{i} = Estep(hme.children{i},x,y);
    end
    
    % evaluate gating function
    switch lower (hme.type)
        case 'linear'
            gate = linearKg_eval(hme.param,x); % bxn matrix
        case 'mlp'
            gate = mlpKg_eval(hme.param,x); % bxn matrix
        case 'localized'
            gate = localizedK_eval(hme.param,[],x); % bxn matrix
    end
    % compute node likelihood and normalized gated child
    % likelihoods
    hme.clik = zeros(size(gate));
    for i = 1:length(hme.children),
      hme.clik(i,:) = gate(i,:) .* hme.children{i}.lik;
    end    
    hme.lik = sum(hme.clik,1);
    for i = 1:length(hme.children),
      hme.clik(i,:) = hme.clik(i,:) ./ (hme.lik+eps);
    end    
  
  end

% end Estep

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% requires hme.{lik,clik} from Estep
% sets hme.param in all nodes
% destroys hme.{lik,clik} in all nodes
function [hme] = Mstep(hme,x,y,w)

  if hme.leaf,

    % train expert
    switch lower(hme.type)
        case 'linear'
            switch lower (hme.optimizer)
                case 'gradiente'
                    [hme.param,hme.sig] = linearGradK(x,y,w,hme.param,hme.sig);
                case 'pseudo_inversao'
                    [hme.param,hme.sig] = linearInversaoK(x,y,w,hme.param,hme.sig);
                otherwise
                    disp(sprintf('Nao implementado o Metodo de treinamento %s',hme.optimizer))
                    pause
            end
        case 'mlp'
            switch lower (hme.optimizer)
                case 'gradiente'
                    [hme.param,hme.sig] = mlpGradK(x,y,w,hme.h,hme.param,hme.sig);
                case 'conjugado'
                    [hme.param,hme.sig] = mlpConjK(x,y,w,hme.h,hme.param,hme.sig);                    
                case 'levenberg'
                    [hme.param,hme.sig] = mlpLevenbergK(x,y,w,hme.h,hme.param,hme.sig);
                otherwise
                    disp(sprintf('Nao implementado Metodo de treinamento %s',hme.optimizer))
                    pause
            end                       
        case 'svm'
            [hme.param,hme.sig] = svmK(x,y,w,hme.param,hme.sig);
        case 'lssvm'
            [hme.param,hme.sig] = lssvmK(x,y,w,hme.param,hme.sig);
    end
  else
    
    % train children
    for i = 1:length(hme.children),
      wi = w .* hme.clik(i,:);
      hme.children{i} = Mstep(hme.children{i},x,y,wi);
    end
    
    % train gating function
    switch lower (hme.type)
        case 'linear' % Modelos lineares
            switch lower (hme.optimizer)
                case 'gradiente'
                    hme.param = linearGradKg(x,hme.clik,w,hme.param);
                case 'pseudo_inversao'
                    hme.param = linearInversaoKg(x,hme.clik,w,hme.param);
                otherwise
                    disp(sprintf('Nao implementado Metodo de treinamento %s',hme.optimizer))
                    pause
            end
        case 'mlp'
            switch lower (hme.optimizer)
                case 'gradiente'
                    hme.param = mlpGradKg(x,hme.clik,w,hme.h,hme.param);
                case 'conjugado'
                    hme.param = mlpConjKg(x,hme.clik,w,hme.h,hme.param);
                case 'levenberg'
                    hme.param = mlpLevenbergKg(x,hme.clik,w,hme.h,hme.param);
                otherwise
                    disp(sprintf('Nao implementado Metodo de treinamento %s',hme.optimizer))
                    pause
            end
        case 'localized'
            hme.param = localizedK(x,hme.clik,w,hme.param);
    end
  end
  
  % blow away stuff from Estep
  if hme.leaf,
    hme = rmfield(hme,{'lik'});
  else
    hme = rmfield(hme,{'lik','clik'});
  end

% end Mstep

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

