function [beta,sig,post,lli] = linearInversaoK(x,y,w,beta,sig)
% [beta,post,lli] = logistK(x,y,beta,w) 
%
% k-class logistic regression with optional sample weights
%
% k = number of classes
% n = number of samples
% d = dimensionality of samples
%
% INPUT
% 	x 	dxn matrix of n input column vectors
% 	y 	kxn vector of class assignments
% 	[w]	1xn vector of sample weights 
%	[beta]	dxk matrix of model coefficients
%
% OUTPUT
% 	beta 	dxk matrix of fitted model coefficients 
%		(beta(:,k) are fixed at 0) 
% 	post 	kxn matrix of fitted class posteriors
% 	lli 	log likelihood
%
% Let p(i,j) = exp(beta(:,j)'*x(:,i)),
% Class j posterior for observation i is:
%	post(j,i) = p(i,j) / (p(i,1) + ... p(i,k))
%
% See also logistK_eval.
%
% David Martin <dmartin@eecs.berkeley.edu> 
% May 3, 2002

% Copyright (C) 2002 David R. Martin <dmartin@eecs.berkeley.edu>
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License as
% published by the Free Software Foundation; either version 2 of the
% License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
% 02111-1307, USA, or see http://www.gnu.org/copyleft/gpl.html.

% TODO - this code would be faster if x were transposed

error(nargchk(3,5,nargin));

debug = 0;
if debug>0,
  h=figure(1);
  set(h,'DoubleBuffer','on');
end

% get sizes
disp('Entrando na funcao linearK')
nx=size(x,2);

x=[x;ones(1,nx)];

[d,nx] = size(x);
[k,ny] = size(y);



% check sizes
%if k < 2,
%  error('Input y must encode at least 2 classes.');
%end
if nx ~= ny,
  error('Inputs x,y not the same length.'); 
end

n = nx;

% make sure class assignments have unit L1-norm
%sumy = sum(y,1);
%if abs(1-sumy) > eps,
%  sumy = sum(y,1);
%  for i = 1:k, y(i,:) = y(i,:) ./ sumy; end
%end
%clear sumy;

% if sample weights weren't specified, set them to 1
if nargin<5
    disp('Utilizando sig = 1')
    sig=1;
    pause
end

if nargin < 3, 
    disp('Utilizando pesos')
  w = ones(1,n);
  pause
end

% normalize sample weights so max is 1
%w = w / max(w);

% if starting beta wasn't specified, initialize randomly
if nargin < 4,
  %beta = 1e-3*rand(d,k);
  beta=rands(d,k)
  %beta(:,k) = 0;	% fix beta for class k at zero
  %else
  %if sum(beta(:,k)) ~= 0,
  %  error('beta(:,k) ~= 0');
  %end
end

iter=1;

post = computePost(beta,x);

lli = computeLogLik(post,y,w,sig);
  
vis(x,y,beta,lli,d,k,iter,debug);

var=sig^2;

W=diag(w/var);

beta=inv(x*W*x')*x*W*y';

post = computePost(beta,x);

lli = computeLogLik(post,y,w,sig);

diff=y-post;

sig=sqrt(sum(w.*(diff.*diff))/sum(w,2));


if debug>0, 
  vis(x,y,beta,lli,d,k,iter,2); 
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% class posteriors
function post = computePost(beta,x)
  [d,n] = size(x);
  [d,k] = size(beta);
  post = zeros(k,n);
  bx = zeros(k,n);
  for j = 1:k, 
    bx(j,:) = beta(:,j)'*x; 
  end
  %for j = 1:k, 
  %  post(j,:) = 1 ./ sum(exp(bx - repmat(bx(j,:),k,1)),1);
  %end
  post=bx;
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% log likelihood
function lli = computeLogLik(post,y,w,sig)
  [k,n] = size(post);
  lli = 0;
  %for j = 1:k,
  %  lli = lli + sum(w.*y(j,:).*log(post(j,:)+eps));
  %end
  var=sig^2; %variancia  
  
  denom=(2*pi*var)^(k/2);  
  
  diff=(y - post);  

  lik=exp(-0.5*(sum(diff.*diff,1))./var)./denom;
  
  lli=sum(w.*log(lik+eps));

  if isnan(lli), 
    error('lli is nan'); 
  end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% gradient and hessian
%% These are computed in what seems a verbose manner, but it is
%% done this way to use minimal memory.  x should be transposed
%% to make it faster.
function [g,h] = derivs(post,x,y,w,sig)

  [k,n] = size(post);
  [d,n] = size(x);

  % first derivative of likelihood w.r.t. beta
  %g = zeros(d,k-1);
  g = zeros(d,k);
  
  var=sig^2;
%  for j = 1:k-1,
  for j = 1:k,
    wyp = w .* (y(j,:) - post(j,:))/var;
    for ii = 1:d, 
      g(ii,j) = x(ii,:) * wyp'; 
    end
  end
  %g = reshape(g,d*(k-1),1);
  g = reshape(g,d*k,1);
  % hessian of likelihood w.r.t. beta
  
  %h = zeros(d*(k-1),d*(k-1)); 
  h = zeros(d*k,d*k); 
    
  %for i = 1:k-1,	% diagonal
  for i = 1:k,	% diagonal
   % wt = w .* post(i,:) .* (1 - post(i,:));
   wt=-w/var;
    hii = zeros(d,d);
    for a = 1:d,
      wxa = wt .* x(a,:);
      for b = a:d,
        hii_ab = wxa * x(b,:)';
        hii(a,b) = hii_ab;
        hii(b,a) = hii_ab;
      end
    end
    h( (i-1)*d+1 : i*d , (i-1)*d+1 : i*d ) = -hii;
  end
  
  %for i = 1:k-1,	% off-diagonal
  %  for j = i+1:k-1,
  %    wt = w .* post(j,:) .* post(i,:);
  %    hij = zeros(d,d);
  %    for a = 1:d,
  %      wxa = wt .* x(a,:);
  %      for b = a:d,
  %        hij_ab = wxa * x(b,:)';
  %        hij(a,b) = hij_ab;
  %        hij(b,a) = hij_ab;
  %      end
  %    end
  %    h( (i-1)*d+1 : i*d , (j-1)*d+1 : j*d ) = hij;
  %    h( (j-1)*d+1 : j*d , (i-1)*d+1 : i*d ) = hij;
  %  end
  %end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% debug/visualization
function vis (x,y,beta,lli,d,k,iter,debug)

  if debug<=0, return, end

  disp(['iter=' num2str(iter) ' lli=' num2str(lli)]);
  if debug<=1, return, end

  if d~=3 | k>10, return, end

  figure(1);
  res = 100;
  r = abs(max(max(x)));
  dom = linspace(-r,r,res);
  [px,py] = meshgrid(dom,dom);
  xx = px(:); yy = py(:);
  points = [xx' ; yy' ; ones(1,res*res)];
  func = zeros(k,res*res);
  for j = 1:k,
    func(j,:) = exp(beta(:,j)'*points);
  end
  [mval,ind] = max(func,[],1);
  hold off; 
  im = reshape(ind,res,res);
  imagesc(xx,yy,im);
  hold on;
  syms = {'w.' 'wx' 'w+' 'wo' 'w*' 'ws' 'wd' 'wv' 'w^' 'w<'};
  for j = 1:k,
    [mval,ind] = max(y,[],1);
    ind = find(ind==j);
    plot(x(1,ind),x(2,ind),syms{j});
  end
  pause(0.1);

% eof
